from lxml import etree
from lxml.html.clean import Cleaner


class DomTreeBuilder(object):
    def __init__(self, html):
        cleaner = Cleaner(style=True, page_structure=False, \
                          remove_tags=('br','img','label','input', 'button', 'link','meta','script','option','selelct',))
        self.html = cleaner.clean_html(html)


    def build(self):
        parser = etree.HTMLParser(encoding='utf-8')
        return etree.fromstring(self.html, parser)


from __future__ import division, print_function

from collections import namedtuple, defaultdict, Counter
import copy
from six import StringIO
from lxml import etree
from lxml.html import tostring, fragment_fromstring
from trees import SimpleTreeMatch, tree_depth, PartialTreeAligner, SimpleTreeAligner, tree_size
from lookupcheck import Scan
import urlparse

GeneralizedNode = namedtuple('GeneralizedNode', ['element', 'length'])
Field = namedtuple('Field', ['text', 'html'])

def element_repr(e):
    return '<%s #%s .%s>' %(e.tag, e.get('class', ''), e.get('id', ''))

def region_to_dict(region):
    return {
        'parent': tostring(region.parent, encoding='unicode', method='html'),
        'start': region.start,
        'k': region.k,
        'covered': region.covered,
        'items': region.items
    }

def dict_to_region(json_region):
    parser = etree.HTMLParser()
    parent = fragment_fromstring(json_region['parent'], parser=parser)
    Region.now=0
    return Region(parent=parent, start=json_region['start'], k=json_region['k'],
                  covered=json_region['covered'],
                  items=json_region['items'])

class Region(object):
    now=0
    url=None
    def __init__(self, now = 0,**dict):
        self.__dict__.update(dict)

    def __str__(self):
        return "<Region: parent {}, start {}:{}, k {}, covered {}>".format(element_repr(self.parent),
                                                                                     self.start,
                                                                                     element_repr(self.parent[self.start]),
                                                                                     self.k,
                                                                                     self.covered)
    __repr__ = __str__

    def __getstate__(self):
        odict = self.__dict__.copy()
        odict['parent'] = tostring(odict['parent'], encoding='unicode', method='html')
        odict['start'] = odict['start']
        odict['k'] = odict['k']
        odict['covered'] = odict['covered']
        odict['items'] = odict['items']
        return odict

    def __setstate__(self, dict):
        parser = etree.HTMLParser(encoding='unicode')
        dict['parent'] = fragment_fromstring(dict['parent'], parser=parser)
        dict['start'] = dict['start']
        dict['k'] = dict['k']
        dict['covered'] = dict['covered']
        dict['items'] = dict['items']
        self.__dict__.update(dict)

    def iter(self, k):
        """
        >>> root = [1, 2, 3, 4, 5]
        >>> region = Region(parent=root, start=1, k=1, covered=4)
        >>> list(region.iter(1))
        [[2], [3], [4], [5]]

        >>> region = Region(parent=root, start=1, k=2, covered=4)
        >>> list(region.iter(2))
        [[2, 3], [4, 5]]

        >>> region = Region(parent=root, start=1, k=1, covered=4)
        >>> list(region.iter(2))
        [[2, 3], [4, 5]]

        """
        for i in range(self.start, self.start + self.covered, k):
            yield self.parent[i:i + k]

    def as_html_table(self, headers=None, show_id=False):
        """
        convert the region to a HTML table
        """
        f = StringIO()
        ff = StringIO()

        print('<table border="1">', file=f)

        # print headers
        if headers:
            print('<tr>', file=f)
            if isinstance(headers, dict):
                if show_id:
                    print('<th></th>', file=f)
                for i in range(len(self.items[0])):
                    print('<th>%s</th>' %headers.get(i, ''), file=f)
            elif isinstance(headers, list):
                if show_id:
                    print('<th></th>', file=f)
                for h in headers:
                    print('<th>%s</th>' %h, file=f)
            print('</tr>', file=f)

        # print content
        score = 0
        index_map = {}
        for i, item in enumerate(self.items):


            print('<tr><td>', file=f)
            temp=[]
            tag_list=[]
            if show_id:
                print('<td>%s</td>' %(i+1), file=f)
            for index,field in enumerate(item):
                #print (field)
                if (Scan(field.text, "Title")):
                    score += 1
                    if index not in index_map:
                        index_map[index] = 0
                        index_map[index] += 1
                data=field.html
                dont=False
                if "tag" in data:
                    tag = data["tag"] if data["tag"] else "div"
                else:
                    tag="div"

                if "attr" in data:
                    #href = data["tag"] if data["tag"] else "div"
                    attr=""
                    for x,y in data["attr"].items():
                        if x=='href' and y!='#':
                            y=urlparse.urljoin(Region.url,y)
                        attr+='%s="%s" '%(x.encode('utf8', 'ignore'),y.encode('utf8', 'ignore'))
                else:
                    attr='class="#"'
                    dont=True

                text=field[0].encode('utf8', 'ignore')
                if not text:
                    if 'href=' in attr:
                        print('<a %s>click here</a>'%attr, file=f)
                        #print(attr, text)
                    if tag_list:
                        if tag_list[-1]=='br':
                            pass
                        else:
                            if len(tag_list)>=6:
                                if tag_list[-6] == 'br':
                                    print('<br>', file=f)
                            else:
                                print('<br>', file=f)
                    else:
                        # print('<br>', file=f)
                        pass
                    tag_list.append('br')
                elif dont :
                    print('%s' %(text), file=f)
                    tag_list.append('n')
                else:
                    tag_list.append('n')
                    if tag=="a":
                        print('<%s %s>%s</%s>' % (tag,attr,text , tag), file=f)
                    else:
                        print('<%s %s>%s' % (tag, attr, text), file=f)
                        temp.insert(0,tag)
            for i in temp:
                print('</%s>'%(i), file=f)
            print('</td></tr>', file=f)

        index_select = sorted(index_map.items(), key=lambda x: x[1], reverse=True)
        selected = index_select[0] if index_select else ""
        print('<th class="score">%s</th>' % str(selected), file=f)
        print('</table>', file=f)
        # print (score,Region.now)

        if Region.now < score:
            print("", file=ff)
            print(f.getvalue(), file=ff)
            # print(self.now,score)
            Region.now = score
            return ff.getvalue()

        print("", file=f)
        return None
        ##print (f.getvalue())

        return f.getvalue()

    def as_html_table1(self, headers=None, show_id=False):
        """
        convert the region to a HTML table
        """
        f = StringIO()
        ff =StringIO()
        print('<table border="1">', file=f)

        # print headers
        if headers:
            print('<tr>', file=f)
            if isinstance(headers, dict):
                if show_id:
                    print('<th></th>', file=f)
                for i in range(len(self.items[0])):
                    print('<th>%s</th>' %headers.get(i, ''), file=f)
            elif isinstance(headers, list):
                if show_id:
                    print('<th></th>', file=f)
                for h in headers:
                    print('<th>%s</th>' %h, file=f)
            print('</tr>', file=f)

        # print content
        score = 0
        index_map={}
        for i, item in enumerate(self.items):
            #print(item)
            print('<tr>', file=f)
            print('<td>', file=f)
            if show_id:
                print('<td>%s</td>' %(i+1), file=f)
            for index,field in enumerate(item):

                # if field.html or field.text:
                        # if field.html:
                        #     #print (field.html)
                        #     print('<td>%s</td>' % field.html, file=f)

                        # if field.text:
                tag=field.html if field.html else "div"
                print('<%s>%s</%s>'%(tag,field[0].encode('utf8', 'ignore'),tag), file=f)

                if (Scan(field.text,"Title")):
                    score +=1
                    if index not in index_map:
                        index_map[index]=0
                        index_map[index]+=1

            print('</td>', file=f)
            print('</tr>', file=f)
        #print(score)
        index_select=sorted(index_map.items(),key=lambda x:x[1],reverse=True)
        selected=index_select[0] if index_select else ""
        print('<th class="score">%s</th>'%str(selected), file=f)
        print('</table>', file=f)

        if Region.now < score:
            print("", file=ff)
            print (f.getvalue(),file=ff)
            #print(self.now,score)
            Region.now=score
            return ff.getvalue()

        print("", file=f)
        return None

    def as_plain_texts(self):
        """
        convert the region to a two dim plain texts.
        """
        return [[field[0] for field in item] for item in self.items]

class Record(object):
    def __init__(self, *elements):
        self.elements = elements

    def __len__(self):
        return len(self.elements)

    def __str__(self):
        return 'DataRecord: %s' % ", ".join(element_repr(e) for e in self.elements)

    def __iter__(self):
        return iter(self.elements)

    def __getitem__(self, item):
        return self.elements[item]

    @staticmethod
    def size(record):
        s = 0
        for element in record.elements:
            s += tree_size(element)
        return s

def pairwise(a, K, start=0):
    """
    A generator to return the comparison pair.

    for example:
    >>> list(pairwise([1, 2, 3, 4], 1))
    [([1], [2]), ([2], [3]), ([3], [4])]
    >>> list(pairwise([1, 2, 3, 4], 2))
    [([1], [2]), ([2], [3]), ([3], [4]), ([2], [3]), ([3], [4]), ([1, 2], [3, 4])]
    >>> list(pairwise([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 2))
    [([1], [2]), ([2], [3]), ([3], [4]), ([4], [5]), ([5], [6]), ([6], [7]), ([7], [8]), ([8], [9]), ([9], [10]), ([2], [3]), ([3], [4]), ([4], [5]), ([5], [6]), ([6], [7]), ([7], [8]), ([8], [9]), ([9], [10]), ([1, 2], [3, 4]), ([3, 4], [5, 6]), ([5, 6], [7, 8]), ([7, 8], [9, 10]), ([2, 3], [4, 5]), ([4, 5], [6, 7]), ([6, 7], [8, 9])]
    >>> list(pairwise([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], 3, 1))
    [([2], [3]), ([3], [4]), ([4], [5]), ([5], [6]), ([6], [7]), ([7], [8]), ([8], [9]), ([9], [10]), ([3], [4]), ([4], [5]), ([5], [6]), ([6], [7]), ([7], [8]), ([8], [9]), ([9], [10]), ([4], [5]), ([5], [6]), ([6], [7]), ([7], [8]), ([8], [9]), ([9], [10]), ([2, 3], [4, 5]), ([4, 5], [6, 7]), ([6, 7], [8, 9]), ([3, 4], [5, 6]), ([5, 6], [7, 8]), ([7, 8], [9, 10]), ([4, 5], [6, 7]), ([6, 7], [8, 9]), ([2, 3, 4], [5, 6, 7]), ([5, 6, 7], [8, 9, 10]), ([3, 4, 5], [6, 7, 8]), ([4, 5, 6], [7, 8, 9])]
    """
    for k in range(1, K + 1):
        for i in range(0, K):
            for j in range(start+i, len(a), k):
                slice_a = a[j:j + k]
                slice_b = a[j + k: j + 2 * k]

                if len(slice_a) >= k and len(slice_b) >= k:
                    yield slice_a, slice_b

class MiningDataRegion(object):
    def __init__(self, root, max_generalized_nodes, threshold):
        self.root = root
        self.max_generalized_nodes = max_generalized_nodes
        self.threshold = threshold
        self.stm = SimpleTreeMatch()

    def find_regions(self, root):
        data_regions = []
        if tree_depth(root) >= 1:
            scores = self.compare_generalized_nodes(root, self.max_generalized_nodes)
            data_regions.extend(self.identify_regions(0, root, self.max_generalized_nodes, self.threshold, scores))
            covered = set()
            for data_region in data_regions:
                for i in range(data_region.start, data_region.covered):
                    covered.add(data_region.parent[i])

            for child in root:
                if child not in covered:
                    data_regions.extend(self.find_regions(child))
        # print (data_regions)
        return data_regions


    def identify_regions(self, start, root, max_generalized_nodes, threshold, scores):
        cur_region = Region(parent=root, start=0, k=0, covered=0, score=0)
        max_region = Region(parent=root, start=0, k=0, covered=0, score=0)
        data_regions = []

        for k in range(1, max_generalized_nodes + 1):
            for i in range(0, max_generalized_nodes):
                flag = True
                for j in range(start + i, len(root) - k, k):
                    pair = GeneralizedNode(root[j], k), GeneralizedNode(root[j + k], k)
                    score = scores.get(pair)
                    if score and score >= threshold:
                        if flag:
                            cur_region.k = k
                            cur_region.start = j
                            cur_region.score = score
                            cur_region.covered = 2 * k
                            flag = False
                        else:
                            cur_region.covered += k
                            cur_region.score += score
                    elif not flag:  # doesn't match but previous match
                        break

                if self.calculate_score(cur_region) > self.calculate_score(max_region):
                    max_region.k = cur_region.k
                    max_region.start = cur_region.start
                    max_region.covered = cur_region.covered
                    max_region.score = cur_region.score

        if max_region.covered:
            data_regions.append(max_region)
            if max_region.start + max_region.covered < len(max_region.parent):
                data_regions.extend(self.identify_regions(max_region.start + max_region.covered, root,
                                                          max_generalized_nodes, threshold, scores))

        return data_regions


    def compare_generalized_nodes(self, parent, k):
        """
         compare the adjacent children generalized nodes similarity of a given element

         Arguments:
         `parent`: the lxml element to compare children of.
         `k`: the maximum length of generalized node.
        """
        scores = {}
        for a, b in pairwise(parent, k):
            gn1 = GeneralizedNode(a[0], len(a))
            gn2 = GeneralizedNode(b[0], len(b))
            if (gn1, gn2) not in scores:
                score = self.stm.normalized_match_score(a, b)
                scores.setdefault((gn1, gn2), score)
        return scores

    def calculate_score(self, region):
        if region.covered == 0:
            return 0
        count = region.covered / region.k
        return region.score / count

class MiningDataRecord(object):
    """
    mining the data record from a region.

    basic assumption:
    the subtree of data records also similar. so if not any adjacent pair of them are
    similar (less than threshold), data region itself is a data record,
    otherwise children are individual data record.
    """

    def __init__(self, threshold):
        self.stm = SimpleTreeMatch()
        self.threshold = threshold

    def find_records(self, region):
        if region.k == 1:
            records = []
            # if all the individual node of children node of Generalized node are similar
            for i in range(region.start, region.start + region.covered):
                for child1, child2 in pairwise(region.parent[i], 1, 0):
                    sim = self.stm.normalized_match_score(child1, child2)
                    if sim < self.threshold:
                        return self.slice_region(region)
            else:
                # each child of generalized node is a data record
                for gn in region.iter(1):
                    records.extend([Record(c) for c in gn])
            return records
        else:
            # if almost all the individual node in Generalized Node are similar
            children = [region.parent[region.start + i] for i in range(region.covered)]
            sizes = Counter([tree_size(child) for child in children])
            most_common_size, _= sizes.most_common(1)[0]
            most_typical_child = [child for child in children if tree_size(child) == most_common_size][0]
            similarities = dict([child, self.stm.normalized_match_score([most_typical_child], [child])] for child in children)
            if self.almost_similar(list(similarities.values()), self.threshold):
                return [Record(child) for child in children if similarities[child] >= self.threshold]
            else:
                return self.slice_region(region)

    def slice_region(self, region):
        """
        slice every generalized node of region to a data record
        """
        records = []
        for gn in region.iter(region.k):
            elements = [element for element in gn]
            records.append(Record(*elements))
        return records

    def almost_similar(self, similarities, threshold):
        sims = [1 for sim in similarities if sim >= threshold]
        return len(sims) / len(similarities) > 0.8

class MiningDataField(object):
    """
    Mining the data item from data records with partial tree alignment.
    """
    def __init__(self, include_html=False):
        self.pta = PartialTreeAligner(SimpleTreeAligner())
        self.sta = SimpleTreeAligner()

    def align_records(self, records):
        """partial align multiple records.

        for example (from paper Web Data Extraction Based on Partial Tree Alignment):
        >>> from lxml.html import fragment_fromstring
        >>> t1 = fragment_fromstring("<p> <x1></x1> <x2></x2> <x3></x3> <x></x> <b></b> <d></d> </p>")
        >>> t2 = fragment_fromstring("<p> <b></b> <n></n> <c></c> <k></k> <g></g> </p>")
        >>> t3 = fragment_fromstring("<p> <b></b> <c></c> <d></d> <h></h> <k></k> </p>")
        >>> mdf = MiningDataField()
        >>> _, seed = mdf.align_records([Record(t1), Record(t2), Record(t3)])
        >>> [e.tag for e in seed[0]]
        ['x1', 'x2', 'x3', 'x', 'b', 'n', 'c', 'd', 'h', 'k', 'g']
        >>> [e.tag for e in t1]
        ['x1', 'x2', 'x3', 'x', 'b', 'd']
        """

        # sort by the tree size
        sorted_records = sorted(records, key=Record.size)

        # seed is the largest tree
        seed = sorted_records.pop()

        # a dict like {'t2': {}, 't3': {}, etc}
        # the nested dictionary is like {'seed_element' : 'original_element'}
        mappings = defaultdict(dict)
        seed_copy = copy.deepcopy(seed)
        mappings.setdefault(seed, self._create_seed_mapping(seed_copy, seed))

        R = []
        items = []
        while len(sorted_records):
            next = sorted_records.pop()
            modified, partial_match, aligned = self.pta.align(seed_copy, next)
            mappings.update({next: aligned})

            if modified:
                sorted_records.extend(R)
                R = []
            else:
                # add it back to try it later since seed might change
                if partial_match:
                    R.append(next)

        for record in records:
            aligned = mappings[record]
            items.append(self._extract_item(seed_copy, aligned))

        return items, seed_copy

    def align_record(self, seed, record):
        """simple align the given record with given seed
        """
        alignment = self.sta.align(seed, record)
        aligned = dict({alignment.first: alignment.second})
        for sub in alignment.subs:
            aligned.update({sub.first: sub.second})
        return self._extract_item(seed, aligned)

    def _create_seed_mapping(self, seed, record):
        """create a mapping from seed record to data record.

        for example:
        >>> from lxml.html import fragment_fromstring
        >>> t1 = fragment_fromstring("<p id='1'> <a></a> <b></b> </p>")
        >>> d1 = Record(t1)
        >>> p1 = t1

        >>> t2 = fragment_fromstring("<p id='2'> <a></a> <b></b> </p>")
        >>> d2 = Record(t2)
        >>> p2 = t2

        >>> mdr = MiningDataField()
        >>> d = mdr._create_seed_mapping(d1, d2)
        >>> d[p1] == p2
        True

        """
        d = {}
        for s, e in zip(seed, record):
            d[s] = e
            d.update(self._create_seed_mapping(s, e))
        return d

    def _extract_item(self, seed, mapping):
        """extract data item from the tree.

        seed: the seed tree
        mapping: a seed element to original element dict
        """
        r = []
        for element in seed:
            r.extend(self._extract_element(element, mapping))
        return r

    def _extract_element(self, seed, mapping):
        r = []
        e = mapping.get(seed, None)

        # handle text
        attributes={}
        tag=None

        if e is not None:
            attributes=e.attrib
            tag=e.tag

        pass_data={"tag":tag,"attr":attributes}

        # if attributes:
        #     href=attributes.get('href')
        #
        #     # if href:
        #     #     r.append(Field(href, b''))
        #         #print (href)

        if e is not None:

            if seed.text and seed.text.strip():
                r.append(Field(self._get_text(e.text),pass_data))
            else:
                r.append(Field(u'', pass_data))
        else:
            if seed.text and seed.text.strip():
                r.append(Field(u'',pass_data))
            else:
                r.append(Field(u'', pass_data))

        # handle children
        for child in seed:
            r.extend(self._extract_element(child, mapping))

        # handle tail
        if e is not None:
            if seed.tail and seed.tail.strip():
                r.append(Field(self._get_text(e.tail) or u'', b''))
        else:
            if seed.tail and seed.tail.strip():
                r.append(Field(u'', b''))

        return r

    def _get_text(self, text):
        if text is not None:
            if isinstance(text, bytes):
                return text.decode('utf8', 'ignore')
            return text
        return u''



cimport cython

def _get_root(e):
    return e.tag if e is not None else None

def _get_child(e, i):
    return e[i]

def _get_children_count(e):
    return len(e)

@cython.boundscheck(False)
@cython.wraparound(False)
def tree_match(t1, t2):

    if _get_root(t1) is None or _get_root(t2) is None:
        return 0
    if _get_root(t1) != _get_root(t2):
        return 0

    cdef int rows = _get_children_count(t1) + 1
    cdef int cols = _get_children_count(t2) + 1
    cdef int i
    cdef int j

    m = create_2d_matrix(rows, cols)

    for i from 1 <= i < rows:
        for j from 1 <= j < cols:
            m[i][j] = max(m[i][j - 1], m[i - 1][j])
            m[i][j] = max(m[i][j], m[i - 1][j - 1] + tree_match(_get_child(t1, i - 1), _get_child(t2, j - 1)))

    return 1 + m[rows-1][cols-1]


@cython.boundscheck(False)
@cython.wraparound(False)
def create_2d_matrix(rows, cols):
    return [[0 for _ in range(cols)] for _ in range(rows)]



__author__ = 'schinnasamy'
import pickle
import os
from django.conf import settings
from urlparse import urljoin
#from django.conf.settings import PROJECT_ROOT

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))
path=PROJECT_ROOT+'/Lookup_data_new_v2.pickle'
path=path.replace('\\','/')

with open(path, 'r') as f:
    lookup_model = pickle.load(f)


def Scan(text, tag):
    # print text
    mNgramLst = lookup_model[0]
    mUnigramLst = lookup_model[1]

    Punc_For_Split = [',', '.', ':', '\t', '-']
    for pun in Punc_For_Split:
        text = text.replace(pun, " %s " % pun)
    # print text
    tmpTokens = text.split(" ")
    # track=0
    MatchedString = ""
    for i in range(0, len(tmpTokens)):
        if tmpTokens[i].lower() in mUnigramLst[tag]:
            if not MatchedString:
                list_c = mUnigramLst[tag][tmpTokens[i].lower()]
                List_f = {x: list_c.count(x) for x in list_c}
                list_s = List_f.keys()
                list_r = reversed(list_s)
                for j in list_r:
                    if len(tmpTokens[i:]) >= j:
                        valid_con = " ".join(tmpTokens[i:j + i])
                        if valid_con.lower() in mNgramLst[tag]:
                            MatchedString = valid_con
                            break


    return MatchedString



from setuptools import setup, find_packages
from distutils.extension import Extension

    
try:
    from Cython.Distutils import build_ext
except ImportError:
    use_cython = False
else:
    use_cython = True

cmdclass = {}
ext_modules = []

##if use_cython:
##    ext_modules.append(Extension("joblist.trees_cython", ['joblist/trees_cython.pyx']))
##    cmdclass.update({'build_ext': build_ext})
##else:
##    ext_modules.append(Extension("joblist.trees_cython", ['joblist/trees_cython.c']))


setup(name='joblist',
      version='2.0.3',
      description="A Python implementation of joblist",
      long_description="A Python implementation of JobList (Data Extraction with Partial Tree Alignment)",
      author="Sellamani",
      author_email="mail2.sella@gmail.com",
      install_requires=['w3lib'],
      packages=find_packages(),
      include_package_data=True,
      #package_data={'joblist': ['joblist/Lookup_data_new_v2.pickle']},
      package_data={'joblist': ['*.pickle', 'joblist/Lookup_data_new_v2.pickle']},
      cmdclass=cmdclass,
      ext_modules=ext_modules
)



from __future__ import division
import copy
import os
import sys
import imp



from trees_cython import create_2d_matrix, tree_match
    

def tree_size(root):
    """
    Get the subtree size.

    for example:
    >>> from lxml.html import fragment_fromstring

    >>> root = fragment_fromstring("<p> </p>")
    >>> tree_size(root)
    1

    >>> root = fragment_fromstring("<p> <font> <a href='#top'> </a> </font> </p>")
    >>> tree_size(root)
    3

    >>> html = "<p><font><font>^ </font><a href='#top'><font>Return to the top</font></a></font><br><font>--</font> </p>"
    >>> root = fragment_fromstring(html)
    >>> tree_size(root)
    7
    """
    if len(root) == 0:
        return 1
    return sum([tree_size(child) for child in root]) + 1

def tree_depth(root):
    """
    Get the subtree depth.

    for example:
    >>> from lxml.html import fragment_fromstring

    >>> root = fragment_fromstring("<p> </p>")
    >>> tree_depth(root)
    1

    >>> root = fragment_fromstring("<p> <font> </font> </p>")
    >>> tree_depth(root)
    2

    >>> html = "<p><font><font>^ </font><a href='#top'><font>Return to the top</font></a></font><br><font>--</font> </p>"
    >>> root = fragment_fromstring(html)
    >>> tree_depth(root)
    4

    """
    if len(root) == 0:
        return 1
    return max([tree_depth(child) for child in root]) + 1

def _get_root(e):
    return e.tag if e is not None else None

def _get_child(e, i):
    return e[i] if i < len(e) else None

def _get_children_count(e):
    return len(e)

class SimpleTreeMatch(object):
    """
    Simple Tree Match.
    """
    def __init__(self, ger_root=_get_root, get_children_count=_get_children_count, get_child=_get_child):
        self.get_root = ger_root
        self.get_children_count = get_children_count
        self.get_child = get_child

    def match(self, l1, l2):
        """
        match two trees list.
        """
        matrix = create_2d_matrix(len(l1) + 1, len(l2) + 1)
        for i in range(1, len(matrix)):
            for j in range(1, len(matrix[0])):
                matrix[i][j] = max(matrix[i][j - 1], matrix[i - 1][j])
                matrix[i][j] = max(matrix[i][j], matrix[i - 1][j - 1] + self._single_match(l1[i - 1], l2[j - 1]))
        return matrix[i][j]

    def normalized_match_score(self, t1, t2):
        t1size = sum([tree_size(e) for e in t1]) + 1
        t2size = sum([tree_size(e) for e in t2]) + 1
        return self.match(t1, t2) / ((t1size + t2size) / 2)

    def _single_match(self, t1, t2):
        return tree_match(t1, t2)

class TreeAlignment(object):

    TRACE_LEFT = 1
    TRACE_UP = 2
    TRACE_DIAG = 3

    def __init__(self, first=None, second=None, score=0):
        self.first = first
        self.second = second
        self.score = score
        self.subs = []

    def add(self, alignment):
        if self.first is None and self.second is None:
            self.first, self.second = alignment.first, alignment.second
        else:
            self.subs.append(alignment)
        self.subs.extend(alignment.subs)

    @property
    def tag(self):
        assert self.first.tag == self.second.tag
        return self.first.tag

    def __str__(self):
        return '{} {}: {}'.format(self.first, self.subs, self.score)


class SimpleTreeAligner(object):
    """
    Simple Tree Aligner.
    """
    def __init__(self, ger_root=_get_root, get_children_count=_get_children_count, get_child=_get_child):
        self.get_root = ger_root
        self.get_children_count = get_children_count
        self.get_child = get_child

    def align(self, l1, l2):
        """
        align 2 lists of trees

        for example:
        >>> get_root = lambda x: x[0]
        >>> get_children_count = lambda x: len(x) - 1
        >>> get_child = lambda x, i: x[i+1]
        >>> sta = SimpleTreeAligner(get_root, get_children_count, get_child)

        >>> l1 = [('a', ('b',), ('c',), ('e',))]
        >>> l2 = [('a', ('b',), ('c',), ('d',))]
        >>> alignment = sta.align(l1, l2)
        >>> get_root(alignment.first)
        'a'
        >>> alignment.score
        3
        >>> [get_root(align.first) for align in alignment.subs]
        ['c', 'b']
        """
        alignment = TreeAlignment()
        matrix = create_2d_matrix(len(l1) + 1, len(l2) + 1)
        alignment_matrix = create_2d_matrix(len(l1), len(l2))
        trace = create_2d_matrix(len(l1), len(l2))

        for i in range(1, len(matrix)):
            for j in range(1, len(matrix[0])):
                if matrix[i][j-1] > matrix[i-1][j]:
                    matrix[i][j] = matrix[i][j-1]
                    trace[i-1][j-1] = TreeAlignment.TRACE_LEFT
                else:
                    matrix[i][j] = matrix[i-1][j]
                    trace[i-1][j-1] = TreeAlignment.TRACE_UP

                alignment_matrix[i-1][j-1] = self.single_align(l1[i - 1], l2[j - 1])
                score = matrix[i-1][j-1] + alignment_matrix[i-1][j-1].score

                if score > matrix[i][j]:
                    matrix[i][j] = score
                    trace[i-1][j-1] = TreeAlignment.TRACE_DIAG

        row = len(trace) - 1
        col = len(trace[0]) - 1

        while row >= 0 and col >= 0:
            if trace[row][col] == TreeAlignment.TRACE_DIAG:
                alignment.add(alignment_matrix[row][col])
                row -= 1
                col -= 1
            elif trace[row][col] == TreeAlignment.TRACE_UP:
                row -= 1
            elif trace[row][col] == TreeAlignment.TRACE_LEFT:
                col -= 1

        alignment.score = matrix[len(matrix)-1][len(matrix[0]) - 1]
        return alignment

    def single_align(self, t1, t2):
        """
        match two single trees.
        >>> get_root = lambda x: x[0]
        >>> get_children_count = lambda x: len(x) - 1
        >>> get_child = lambda x, i: x[i+1]
        >>> sta = SimpleTreeAligner(get_root, get_children_count, get_child)

        >>> t1 = ('a', ('b',), ('c',), ('e',))
        >>> t2 = ('a', ('b',), ('c',), ('d',))
        >>> alignment = sta.single_align(t1, t2)
        >>> (get_root(alignment.first), get_root(alignment.second))
        ('a', 'a')
        >>> alignment.score
        3
        >>> [get_root(align.first) for align in alignment.subs]
        ['c', 'b']

        >>> t1 = ('c', ('b',), ('e',))
        >>> t2 = ('c', ('b',), ('e',))
        >>> alignment = sta.single_align(t1, t2)
        >>> alignment.score
        3
        >>> [get_root(align.first) for align in alignment.subs]
        ['e', 'b']

        >>> t1 = ('a', ('b',), ('c', ('b',), ('e',)), ('e',))
        >>> t2 = ('a', ('b',), ('c', ('b',), ('e',)), ('d',))
        >>> alignment = sta.single_align(t1, t2)
        >>> alignment.score
        5
        >>> [get_root(align.first) for align in alignment.subs]
        ['c', 'e', 'b', 'b']

        >>> t1 = ('a', ('b',), ('c', ('b', ('a',)), ('e',)), ('e',))
        >>> t2 = ('a', ('b',), ('c', ('b', ('a',)), ('e',)), ('d',))
        >>> alignment = sta.single_align(t1, t2)
        >>> alignment.score
        6
        >>> [get_root(align.first) for align in alignment.subs]
        ['c', 'e', 'b', 'a', 'b']

        >>> t1 = ('p', ('a',), ('b',), ('e',))
        >>> t2 = ('p', ('b',), ('c',), ('d',), ('e',))

        >>> alignment = sta.single_align(t1, t2)
        >>> alignment.score
        3

        >>> from lxml.html import fragment_fromstring, tostring
        >>> t1 = fragment_fromstring("<div> <h1></h1> <h2></h2> <h5></h5> </div>")
        >>> t2 = fragment_fromstring("<div> <h2></h2> <h3></h3> <h4></h4> <h5></h5> </div>")

        >>> (_get_root(t1), _get_children_count(t1), _get_child(t1, 1).tag)
        ('div', 3, 'h2')

        >>> (_get_root(t2), _get_children_count(t2), _get_child(t2, 1).tag)
        ('div', 4, 'h3')

        >>> sta = SimpleTreeAligner()
        >>> alignment = sta.single_align(t1, t2)
        >>> alignment.score
        3

        """
        if self.get_root(t1) is None or self.get_root(t2) is None:
            return TreeAlignment()
        if self.get_root(t1) != self.get_root(t2):
            return TreeAlignment()

        alignment = TreeAlignment(t1, t2)

        t1_len = self.get_children_count(t1)
        t2_len = self.get_children_count(t2)

        matrix = create_2d_matrix(t1_len + 1, t2_len + 1)
        alignment_matrix = create_2d_matrix(t1_len, t2_len)
        trace = create_2d_matrix(t1_len, t2_len)

        for i in range(1, len(matrix)):
            for j in range(1, len(matrix[0])):
                if matrix[i][j - 1] > matrix[i - 1][j]:
                    matrix[i][j] = matrix[i][j - 1]
                    trace[i - 1][j - 1] = TreeAlignment.TRACE_LEFT
                else:
                    matrix[i][j] = matrix[i - 1][j]
                    trace[i - 1][j - 1] = TreeAlignment.TRACE_UP
                alignment_matrix[i - 1][j - 1] = self.single_align(self.get_child(t1, i - 1), self.get_child(t2, j - 1))
                score = matrix[i - 1][j - 1] + alignment_matrix[i - 1][j - 1].score
                if score > matrix[i][j]:
                    matrix[i][j] = score
                    trace[i - 1][j - 1] = TreeAlignment.TRACE_DIAG

        row = len(trace) - 1

        if row >= 0:
            col = len(trace[0]) - 1
        else:
            col = -1

        while row >= 0 and col >= 0:
            if trace[row][col] == TreeAlignment.TRACE_DIAG:
                alignment.add(alignment_matrix[row][col])
                row -= 1
                col -= 1
            elif trace[row][col] == TreeAlignment.TRACE_UP:
                row -= 1
            elif trace[row][col] == TreeAlignment.TRACE_LEFT:
                col -= 1

        alignment.score = 1 + matrix[len(matrix) - 1][len(matrix[0]) - 1]
        return alignment

def find_subsequence(iterable, predicate):
    """
    find the subsequence in iterable which predicted return true

    for example:
    >>> find_subsequence([1, 10, 2, 3, 5, 8, 9],  lambda x: x in [1, 0, 2, 3, 4, 5, 8, 9])
    [[1], [2, 3, 5, 8, 9]]

    >>> find_subsequence([1, 10, 2, 3, 5, 8, 9],  lambda x: x not in [1, 0, 2, 3, 4, 5, 8, 9])
    [[10]]

    """
    seqs = []
    seq = []
    continuous = False
    for i in iterable:
        if predicate(i):
            if continuous:
                seq.append(i)
            else:
                seq = [i]
                continuous = True
        elif continuous:
            seqs.append(seq)
            seq = []
            continuous = False
    if len(seq):
        seqs.append(seq)
    return seqs

class PartialTreeAligner(object):
    def __init__(self, aligner, **options):
        self.sta = aligner
        self.options = options

    def align(self, l1, l2):
        """
        partial align lxml tree list (l2) to another lxml tree list (l1).

        for example (from Web Data Extraction Based on Partial Tree Alignment):
        >>> from lxml.html import fragment_fromstring, tostring
        >>> sta = SimpleTreeAligner()
        >>> pta = PartialTreeAligner(sta)

        "flanked by 2 sibling nodes"
        >>> t1 = fragment_fromstring("<p> <a></a> <b></b> <e></e> </p>")
        >>> t2 = fragment_fromstring("<p> <b></b> <c></c> <d></d> <e></e> </p>")
        >>> _, _, mapping = pta.align([t1], [t2])
        >>> [e.tag for e in t1]
        ['a', 'b', 'c', 'd', 'e']
        >>> sorted([e.tag for e in mapping.itervalues()])
        ['b', 'c', 'd', 'e', 'p']

        "rightmost nodes"
        >>> t1 = fragment_fromstring("<p> <a></a> <b></b> <e></e> </p>")
        >>> t2 = fragment_fromstring("<p> <e></e> <f></f> <g></g> </p>")
        >>> _, _, mapping = pta.align([t1], [t2])
        >>> [e.tag for e in t1]
        ['a', 'b', 'e', 'f', 'g']
        >>> sorted([e.tag for e in mapping.itervalues()])
        ['e', 'f', 'g', 'p']

        "leftmost nodes"
        >>> t1 = fragment_fromstring("<p> <a></a> <b></b> <e></e> </p>")
        >>> t2 = fragment_fromstring("<p> <f></f> <g></g> <a></a> </p>")
        >>> _, _, mapping = pta.align([t1], [t2])
        >>> [e.tag for e in t1]
        ['f', 'g', 'a', 'b', 'e']
        >>> sorted([e.tag for e in mapping.itervalues()])
        ['a', 'f', 'g', 'p']

        "no unique insertion"
        >>> t1 = fragment_fromstring("<p> <a></a> <b></b> <e></e> </p>")
        >>> t2 = fragment_fromstring("<p> <a></a> <g></g> <e></e> </p>")
        >>> _, _, mapping = pta.align([t1], [t2])
        >>> [e.tag for e in t1]
        ['a', 'b', 'e']
        >>> sorted([e.tag for e in mapping.itervalues()])
        ['a', 'e', 'p']

        "multiple unaligned nodes"
        >>> t1 = fragment_fromstring("<p> <x></x> <b></b> <d></d> </p>")
        >>> t2 = fragment_fromstring("<p> <b></b> <c></c> <d></d> <h></h> <k></k> </p>")
        >>> _, _, mapping = pta.align([t1], [t2])
        >>> [e.tag for e in t1]
        ['x', 'b', 'c', 'd', 'h', 'k']
        >>> sorted([e.tag for e in mapping.itervalues()])
        ['b', 'c', 'd', 'h', 'k', 'p']

        """
        alignment = self.sta.align(l1, l2)
        aligned = dict({alignment.first: alignment.second})
        for sub in alignment.subs:
            aligned.update({sub.first: sub.second})

        # add reverse mapping too
        reverse_aligned = dict(reversed(i) for i in list(aligned.items()))

        modified = False

        unaligned_elements = self.find_unaligned_elements(aligned, l2)
        for l in unaligned_elements:
            left_most = l[0]
            right_most = l[-1]

            prev_sibling = left_most.getprevious()
            next_sibling = right_most.getnext()

            if prev_sibling is None:
                if next_sibling is not None:
                    # leftmost alignment
                    next_sibling_match = reverse_aligned.get(next_sibling, None)
                    for i, element in enumerate(l):
                        element_copy = copy.deepcopy(element)
                        next_sibling_match.getparent().insert(i, element_copy)
                        aligned.update({element_copy: element})
                    modified = True

            elif next_sibling is None:
                # rightmost alignment
                prev_sibling_match = reverse_aligned.get(prev_sibling, None)
                previous_match_index = self._get_index(prev_sibling_match)
                # unique insertion
                for i, element in enumerate(l):
                    element_copy = copy.deepcopy(element)
                    prev_sibling_match.getparent().insert(previous_match_index + 1 + i, element_copy)
                    aligned.update({element_copy: element})
                modified = True
            else:
                # flanked by two sibling elements
                prev_sibling_match = reverse_aligned.get(prev_sibling, None)
                next_sibling_match = reverse_aligned.get(next_sibling, None)

                if prev_sibling_match is not None and next_sibling_match is not None:
                    next_match_index = self._get_index(next_sibling_match)
                    previous_match_index = self._get_index(prev_sibling_match)
                    if next_match_index - previous_match_index == 1:
                        # unique insertion
                        for i, element in enumerate(l):
                            element_copy = copy.deepcopy(element)
                            prev_sibling_match.getparent().insert(previous_match_index + 1 + i, element_copy)
                            aligned.update({element_copy: element})
                        modified = True
        return modified, len(unaligned_elements) > 0, aligned

    def find_unaligned_elements(self, aligned, elements):
        """
        find the unaligned elements recursively from elements.

        >>> from lxml.html import fragment_fromstring
        >>> t1 = fragment_fromstring("<div> <h1></h1> <h2></h2> <h5></h5> </div>")
        >>> t2 = fragment_fromstring("<div> <h2></h2> <h3></h3> <h4></h4> <h5></h5> <h6></h6></div>")

        >>> sta = SimpleTreeAligner()
        >>> pta = PartialTreeAligner(None)
        >>> aligned = dict((align.first, align.second) for align in sta.align([t1], [t2]).subs)
        >>> unaligned = pta.find_unaligned_elements(aligned, [t2])
        >>> [[e.tag for e in l] for l in unaligned]
        [['h3', 'h4'], ['h6']]
        """
        predicate = lambda x: x not in list(aligned.values())
        unaligned = []

        for element in elements:
            current_level = find_subsequence(element, predicate)
            unaligned.extend(current_level)

            for child in element:
                unaligned.extend(find_subsequence(child, predicate))
        return unaligned

    def _get_index(self, element):
        """
        get the position of the element within the parent.
        """
        return element.getparent().index(element)



from __future__ import print_function

import requests
from w3lib.encoding import html_to_unicode
from htmls import DomTreeBuilder
from mdr import MiningDataRegion, MiningDataRecord, MiningDataField,Region



class Depta(object):
    def __init__(self, threshold=0.7, k=5):
        self.threshold = threshold
        self.k = k
    def extract(self, html='', **kwargs):
        Region.now=0
        
        """
        extract data field from raw html or from a url.
        """
        if not html and 'url' in kwargs:
            url=kwargs.pop('url')
            r = requests.get(url)
            Region.url=url
            _, html = html_to_unicode(r.headers.get('content-type'), r.content)

        builder = DomTreeBuilder(html)
        root = builder.build()

        region_finder = MiningDataRegion(root, self.k, self.threshold)
        regions = region_finder.find_regions(root)

        record_finder = MiningDataRecord(self.threshold)
        field_finder = MiningDataField()

        for region in regions:
            #print (region)
            records = record_finder.find_records(region)
            items, _ = field_finder.align_records(records)
            region.items = items
            # print (items)
            if 'verbose' in kwargs:
                print(region)
                for record in records:
                    print('\t', record)

        return regions

